import abc
import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def get_default_alert_streams():
    """Return the AlertStreams configured in settings.py
    """
    try:
        alert_streams = get_alert_streams(settings.ALERT_STREAMS)
    except AttributeError as err:
        raise ImproperlyConfigured(err)

    return alert_streams


def get_alert_streams(alert_stream_configs: list):
    """Return the AlertStreams configured in the given alert_stream_configs
    (a list of configuration dictionaries )

    Use get_default_alert_streams() if you want the AlertStreams configured in settings.py.
    """
    alert_streams = []  # build and return this list of AlertStream subclass instances
    for alert_stream_config in alert_stream_configs:
        if not alert_stream_config.get('ACTIVE', True):
            logger.debug(f'get_alert_streams - ignoring inactive stream: {alert_stream_config["NAME"]}')
            continue  # skip configs that are not active; default to ACTIVE
        try:
            klass = import_string(alert_stream_config['NAME'])
        except ImportError:
            msg = (
                f'The module (the value of the NAME key): {alert_stream_config["NAME"]} could not be imported. '
                f'Check your ALERT_STREAMS setting.'
            )
            raise ImproperlyConfigured(msg)

        alert_stream: AlertStream = klass(**alert_stream_config.get("OPTIONS", {}))
        alert_streams.append(alert_stream)

    return alert_streams


class AlertStream(abc.ABC):
    """Base class for specific alert streams like Hopskotch, GCNClassic, etc.

    * kwargs to __init__ is the OPTIONS dictionary defined in ALERT_STREAMS configuration
    dictionary (for example, see settings.py).
    * allowed_keys and required_keys should be defined as class properties in subclasses.
    * The allowed_keys are turned into instance properties in __init__.
    * Missing required_keys result in an ImproperlyConfigured Django exception.


    To implmement subclass:
    1. define allowed_key, required_keys as class variables
       <say what these do: used with OPTIONS dict in ALERT_STREAMS config dict>
    2. implement listen()
       this method probably doesn't return
    3. write your alert_handlers. which proably take and alert do something.
       The HopskotchAlertStream.listen() method defines an 'alert_handlers' dictionary keyed by
       alert topic with callable values (i.e call this method with alerts from this topic).
       The GCNClassicAlertStream.listen() is another example.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

        # filter the kwargs by allowed keys and add them as properties to AlertStream instance
        self.__dict__.update((k.lower(), v) for k, v in kwargs.items() if k in self.allowed_keys)

        missing_keys = set(self.required_keys) - set(kwargs.keys())
        if missing_keys:
            msg = (
                f'The following required keys are missing from the configuration OPTIONS of '
                f'{self._get_stream_classname()}: {list(missing_keys)} ; '
                f'These keys were found: {list(kwargs.keys())} ; '
                f'Check your ALERT_STREAMS setting.'
            )
            raise ImproperlyConfigured(msg)

        self.alert_handler = self._process_topic_handlers()

    def _get_stream_classname(self) -> str:
        return type(self).__qualname__

    def _process_topic_handlers(self):
        """ convert the TOPIC_HANDLERS values in to callable functions in
        the returned message_handler dictionary. (keyed by topic, value is callable)
        """
        alert_handler = {}
        for topic, callable_string in self.topic_handlers.items():
            # convert string from TOPIC_HANDLERS in to a callable function in
            # the message_handler dictionary (both keyed by topic string)
            alert_handler[topic] = import_string(callable_string)
        return alert_handler

    @abc.abstractmethod
    def listen(self):
        """Listen at the steam and dispatch alerts to handlers. Subclass extentions of
        this method are not expected to return. See hopskotch.py and gcn.py for example
        implementations.
        """
        pass



from datetime import datetime, timezone
import logging
import re
import uuid
import traceback

from django.utils import timezone as tz
from django.core.exceptions import ImproperlyConfigured

from hop import Stream
from hop.auth import Auth
from hop.models import JSONBlob
from hop.io import Metadata, StartPosition, list_topics

from tom_alertstreams.alertstreams.alertstream import AlertStream

logger = logging.getLogger(__name__)


class HopskotchAlertStream(AlertStream):
    """
    """
    required_keys = ['URL', 'GROUP_ID', 'USERNAME', 'PASSWORD', 'TOPIC_HANDLERS']
    allowed_keys = ['URL', 'GROUP_ID', 'USERNAME', 'PASSWORD', 'TOPIC_HANDLERS']
    PUBLIC_TOPIC_CHECK_INTERVAL = 300  # Seconds between checking for new public topics
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # the following methods may fail if improperly configured.
        # So, do them now to catch any errors, before listen() is spawned in it's own Process.
        self.public_topics = self.get_all_public_topics()
        self.stream_url = self.get_stream_url()
        self.stream = self.get_stream()

    def get_all_public_topics(self) -> list[str]:
        """Returns the up-to-date list of Topic names to consume.

        Use the saved options to repeatedly construct the topic list, and
        keep it in sync with the publicaly_readable topics from SCiMMA Auth.

        The Topic list is a combination of the
          a. the publicly_readable Topics from SCiMMA Auth
          b. any topics supplied on the command line via -T, --topic
        """
        hop_auth = Auth(self.username, self.password)
        logger.info('getting publicly_readable topics from SCiMMA Auth.')
        # use the hop-client to ask Kafka directly for the topics since SCiMMA Auth can be out of sync
        # include only topics that a) contain a '.'; b) don't start with '__' (excludes __consumer_offsets)
        publicly_readable_topics = [topic for topic in list_topics(self.url, hop_auth).keys()
                                    if not (topic.startswith('__') and (topic.count('.')==0))]
        logger.info(f'publicly_readable_topics: {publicly_readable_topics}')

        return publicly_readable_topics

    def get_stream_url(self) -> str:
        """For Hopskotch, topics are specified on the url. So, this
        method gets a base url (from super) and then adds topics to it.

        Hopskotch (hop.io) requires at least one topic to be specified.

        You might not need a method like this if your Kafka client provides
        alternative ways to subscribe to a topic. For example, the gcn_kafka.Consumer
        class provides a 'substribe([list of topics])' method. (see gcn.py).
        """
        logger.debug(f'HopskotchAlertStream.get_stream_url topics: {list(self.topic_handlers.keys())}')
        if self.topic_handlers == {}:
            msg = 'Hopskotch requires at least one topic to open the stream. Check ALERT_STREAMS in settings.py'
            raise ImproperlyConfigured(msg)

        base_stream_url = self.url

        # if not present, add trailing slash to base_stream url
        # so, comma-separated topics can be appeneded.
        if base_stream_url[-1] != '/':
            base_stream_url += '/'

        # append comma-separated topics to base URL
        specified_topics = list(self.topic_handlers.keys())
        if '*' in specified_topics:
            # Add all public topics if a asterisk is set in the topic_handlers
            specified_topics = list(set(specified_topics + self.public_topics))
        # Also remove topics with wildcards in them
        specified_topics = [topic for topic in specified_topics if not '*' in topic]

        topics = ','.join(specified_topics)  # 'topic1,topic2,topic3'
        hopskotch_stream_url = base_stream_url + topics

        logger.debug(f'HopskotchAlertStream.get_stream_url url: {hopskotch_stream_url}')
        return hopskotch_stream_url

    def get_stream(self, start_position=StartPosition.LATEST) -> Stream:
        hop_auth = Auth(self.username, self.password)

        # TODO: allow StartPosition to be set from OPTIONS configuration dictionary
        stream = Stream(auth=hop_auth, start_at=start_position)
        return stream

    def listen(self):
        super().listen()
        # TODO: alternatively, WARN upon OPTIONS['topics'] extries that don't have
        # handlers in the alert_handler. (i.e they've configured a topic subscription
        # without providing a handler for the topic. So, warn them).
        last_check_time = tz.now()
        while True:
            try:
                logger.info(f'HopskotchAlertStream.listen opening stream: {self.stream_url} with group_id: {self.group_id}')
                with self.stream.open(self.stream_url, 'r', group_id=self.group_id) as src:
                    for alert, metadata in src.read(metadata=True):
                        # type(gcn_circular) is <hop.models.GNCCircular>
                        # type(metadata) is <hop.io.Metadata>
                        if metadata.topic in self.alert_handler:
                            # TODO: should probably use *args, **kwargs to pass unknow number of arguments
                            self.alert_handler[metadata.topic](alert, metadata)
                        elif '*' in self.alert_handler:
                            # First check all wildcard topics to see if they will match this topic
                            matched_handler = False
                            for topic in self.alert_handler.keys():
                                if topic != '*' and '*' in topic and re.match(topic, metadata.topic):
                                    self.alert_handler[topic](alert, metadata)
                                    matched_handler = True
                                    break
                            # If nothing matched, fall back to default public topic handler
                            if not matched_handler:
                                self.alert_handler['*'](alert, metadata)
                        else:
                            logger.error(f'alert from topic {metadata.topic} received but no handler defined.')
                            # TODO: should define a default handler for all unhandeled topics
                        if (tz.now() - last_check_time).total_seconds() > self.PUBLIC_TOPIC_CHECK_INTERVAL:
                            last_check_time = tz.now()
                            public_topics = self.get_all_public_topics()
                            if set(public_topics) != set(self.public_topics):
                                logger.info(f"New public topics found, restarting hop stream")
                                self.public_topics = public_topics
                                self.stream_url = self.get_stream_url()
                                break
            except Exception as ex:
                logger.error(f'HopskotchAlertStream.listen: {ex}')
                logger.error(traceback.format_exc())  # Show the traceback so we have a chance of figuring out what is breaking

def heartbeat_handler(heartbeat: JSONBlob, metadata: Metadata):
    """Example alert handler for HopskotchAlertStream sys.heartbeat topic.

    Note that HopskotchAlertStream.listen() method knows that Hopskotch alerts come with
    both alert and metadata. So, the alert_handler methods have a signiture (taking both
    as arguments) specific to this stream.
    """
    content: dict = heartbeat.content  # see hop_client reatthedocs
    timestamp = datetime.fromtimestamp(content["timestamp"] / 1e6, tz=timezone.utc)
    if heartbeat.content['count'] % 300 == 0:
        # mod 300 just for convenience so as not to flood logger
        logging.info(f'{timestamp.isoformat()} heartbeat.content dict: {heartbeat.content}. metadata: {metadata}')


def alert_logger(alert: JSONBlob, metadata: Metadata):
    """Example alert handler. The method signsture is specific to Hopskotch alerts.
    """
    # search the header (list of tuples) for a UUID-tuple (keyed by '_id')
    # eg. ('_id', b'$\xd6oGmVM\xed\x97\xe7|\x1c\x8f\x11V\xe9')
    alert_uuid_tuple = next((item for item in metadata.headers if item[0] == '_id'), None)
    if alert_uuid_tuple:
        alert_uuid = uuid.UUID(bytes=alert_uuid_tuple[1])
    else:
        # in this case the alert was probably published with hop-client<0.8.0
        alert_uuid = None
    logger.info(f'Alert (uuid={alert_uuid}) received on topic {metadata.topic}: {alert};  metatdata: {metadata}')



import logging

from gcn_kafka import Consumer

from tom_alertstreams.alertstreams.alertstream import AlertStream


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class GCNClassicAlertStream(AlertStream):
    """

    Pre-requisite: visit gcn.nasa.gov and sign-up to get your client_id and
    client_secret.
    """
    # Upon __init__, the AlertStream base class creates instance properties from
    # the settings OPTIONS dictionary, converting the keys to lowercase.
    required_keys = ['GCN_CLASSIC_CLIENT_ID', 'GCN_CLASSIC_CLIENT_SECRET', 'TOPIC_HANDLERS']
    allowed_keys = ['GCN_CLASSIC_CLIENT_ID', 'GCN_CLASSIC_CLIENT_SECRET', 'TOPIC_HANDLERS', 'DOMAIN', 'CONFIG']

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # properties have been created from the OPTIONS dicttionary

    def listen(self):
        super().listen()

        consumer = Consumer(client_id=self.gcn_classic_client_id,
                            client_secret=self.gcn_classic_client_secret,
                            domain=self.domain,
                            config=self.config,
                            )

        consumer.subscribe(list(self.topic_handlers.keys()))

        # logger.debug(f'Here is a list of the available topics for {self.domain}')
        # for topic in consumer.list_topics().topics:
        #     logger.debug(f'topic: {topic}')

        # what is a cimpl.Message?, cimpl.KafkaError?
        # see https://docs.confluent.io/4.1.1/clients/confluent-kafka-python/index.html#message
        while True:
            for alert in consumer.consume():
                kafka_error = alert.error()  # cimpl.KafkaError
                if kafka_error is None:
                    # no error, so call the alert handler
                    topic = alert.topic()
                    try:
                        self.alert_handler[topic](alert)
                    except KeyError as err:
                        logger.error(f'alert from topic {topic} received but no handler defined. err: {err}')
                else:
                    logger.error(f'GCNClassicAlertStream KafkaError: {kafka_error.name()}: {kafka_error.str()}')
        consumer.close()


def alert_logger(alert):
    """Example alert handler for GCN Classic over Kafka

    This alert handler simply logs the topic and value of the cimpl.Message instance.

    See https://docs.confluent.io/4.1.1/clients/confluent-kafka-python/index.html#message
    for cimpl.Message details.
    """
    logger.info(f'gcn.alert_logger alert.topic(): {alert.topic()}')
    logger.info(f'gcn.alert_logger alert.value(): {alert.value()}')




import datetime
import logging

from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand

from tom_alertstreams.alertstreams.alertstream import get_default_alert_streams
from tom_alertstreams.alertstreams.hopskotch import HopskotchAlertStream

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)


class Command(BaseCommand):
    help = 'Publish a timestamped test message to Hopskotch tomtoolkit.test topic.'

    def handle(self, *args, **options):
        logger.debug(f'hoptestpub.Command.handle() args: {args}')
        logger.debug(f'hoptestpub.Command.handle() options: {options}')

        try:
            alert_streams = get_default_alert_streams()
            # extract the HopskotchAlertStream
            hopskotch_alert_stream = next(stream for stream in alert_streams if isinstance(stream, HopskotchAlertStream))

        except ImproperlyConfigured as ex:
            logger.error(f'{ex.__class__.__name__}: Configure alert streams in settings.py ALERT_STREAMS: {ex}')
            exit(1)

        stream = hopskotch_alert_stream.get_stream()
        topic = 'tomtoolkit.test'  # SCiMMA Admin topic permissions for Credential are assumed to have been set up

        with stream.open(hopskotch_alert_stream.url+topic, "w") as s:
            s.write({
                'created': datetime.datetime.utcnow().isoformat(),
                'created-by': 'tom-alertstreams hoptestpub.py'
             })

        logger.info('hoptestpub Command.handle() returning...')



import logging
from threading import Thread

from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand

from tom_alertstreams.alertstreams.alertstream import get_default_alert_streams


logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)


class Command(BaseCommand):
    help = 'Consume alerts from the alert streams configured in the settings.py ALERT_STREAMS'

    def handle(self, *args, **options):
        logger.debug(f'readstreams.Command.handle() args: {args}')
        logger.debug(f'readstreams.Command.handle() options: {options}')

        try:
            alert_streams = get_default_alert_streams()
        except ImproperlyConfigured as ex:
            logger.error(f'{ex.__class__.__name__}: Configure alert streams in settings.py ALERT_STREAMS: {ex}')
            exit(1)

        try:
            # listen to each alert_stream in it's own Thread (sort of at the same time)
            for alert_stream in alert_streams:
                t = Thread(target=alert_stream.listen, name=alert_stream._get_stream_classname())
                t.start()
                logger.info((f'read_streams {alert_stream._get_stream_classname()} TID={t.native_id} ; '
                             f'thread identifier={t.ident}'))
        except KeyboardInterrupt as msg:
            logger.info(f'read_streams handling KeyboardInterupt {msg}')

        logger.info('readstreams Command.handle() returning...')
